#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAIN_C
#include "exitfail.h"

#define SERIAL_BAUDRATE B115200
#define BUF_SIZE 2048

static int serial_fd = -1;
static struct termios old_tio;

static int terminated = 0;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) +1)
#define ARRAY_SIZE(a) (sizeof(a) /sizeof((a)[0]))

#define __unused __attribute__((unused))

/**** sample: response data from nodes: begin ****/
static struct {
	int gid;
	int rid;
	int ch;
	char msg[24 + 1]; /* 24 characters */
	int idx;
	int sid;
	char rt[20 + 1]; /* 20 characters */
} res_data;

static struct {
	char unit_type[2 + 1]; /* 2 characters */
	char ctrl_code[2 + 1]; /* 2 characters */
	char battery[2 + 1]; /* 2 characters */
	char data[16 + 1]; /* 16 characters */
	float tmp;
	float hum;
	int lux;
} msg_03;

static struct {
	char unit_type[2 + 1]; /* 2 characters */
	char ctrl_code[2 + 1]; /* 2 characters */
	char battery[2 + 1]; /* 2 characters */
	char data[12 + 1]; /* 12 characters */
	int detect_count;
} msg_0b;

static struct {
	char unit_type[2 + 1]; /* 2 characters */
	char ctrl_code[2 + 1]; /* 2 characters */
	char sub_code[2 + 1]; /* 2 characters */
	char data[12 + 1]; /* 12 characters */
	int co2;
} msg_20;
/**** sample: response data from nodes: end ****/

/**** sample: Nodes response: begin ****/
static unsigned char res_start[] = "GID"; /* response: data start */
/**** sample: Nodes response: end ****/

static void terminate_sig_handler(__unused int signum)
{
	terminated = 1;
}

static void set_sig_handler(int sig_list[], ssize_t num, __sighandler_t handler)
{
	struct sigaction sa;
	int i;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;

	for(i=0; i<num; i++)
		if(sigaction(sig_list[i], &sa, NULL) < 0)
			exitfail_errno("sigaction");
}

static void restore_serial(void)
{
	int ret;

	ret = tcsetattr(serial_fd, TCSANOW, &old_tio);
	if(ret < 0) exitfail_errno("tcsetattr");
}

static void setup_serial(int fd)
{
	struct termios tio;
	int ret;

	ret = tcgetattr(fd, &old_tio);
	if(ret) exitfail_errno("tcsgetattr");

	if(atexit(restore_serial)) exitfail_errno("atexit");

	memset(&tio, 0, sizeof(tio));
	tio.c_iflag = IGNBRK | IGNPAR;
	tio.c_cflag = CS8 | CLOCAL | CREAD;
	tio.c_cc[VTIME] = 0;
	tio.c_cc[VMIN] = 1;
	ret = cfsetspeed(&tio, SERIAL_BAUDRATE);
	if(ret < 0) exitfail_errno("cfsetspeed");

	ret = tcflush(fd, TCIFLUSH);
	if(ret < 0) exitfail_errno("tcflush");

	ret = tcsetattr(fd, TCSANOW, &tio);
	if(ret) exitfail_errno("tcsetattr");
}

/**** sample: char(HEX) to int: begin ****/ 
int ch2i(unsigned char c) {
	int ret;
	
	ret = 0;

	if(c>='0' && c<='9') ret = (unsigned int)c - (unsigned int)'0';	
	else if(c>='a' && c <='f') ret = (unsigned int)c - (unsigned int)'a' + 10;	
	else if(c>='A' && c <='F') ret = (unsigned int)c - (unsigned int)'A' + 10;
	else ret = -1;

	return ret;	
}
/**** sample: char(HEX) to int: end ****/

int main(int argc, char *argv[])
{
	int terminate_sig_list[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};

/**** sample: variable: begin ****/
	ssize_t ret;
       unsigned char buf[BUF_SIZE];
       unsigned char buf1[BUF_SIZE];
	int i, j;
	int flag_start, flag_end;
	int cnt_cmp;
	FILE *fp;
	ssize_t analen;
/**** sample: variable: end ****/

	exitfail_init();

	if(argc < 2) {
		printf("Usage: %s <device>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	serial_fd = open(argv[1], O_RDWR | O_NOCTTY);
	if(serial_fd < 0) exitfail_errno("open");

	set_sig_handler(terminate_sig_list, ARRAY_SIZE(terminate_sig_list),
			terminate_sig_handler);

	setup_serial(serial_fd);

/**** sample: receive data from nodes: begin ****/
	printf("\n****************************************************\n");
	printf("-------- Start to reveive data from nodes --------\n");

	flag_start = 0;
	flag_end = 0;
	cnt_cmp = 0;
	analen = 0;

	while(!terminated) {
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_end != 1; i++) {
//			printf("%c", buf[i]);
			/* Not considered about search response. ex."GIGID" */
			if(flag_start != 1 && buf[i] == res_start[cnt_cmp]) {
				cnt_cmp++;
				buf1[analen++] = buf[i];
			}
			else {
				cnt_cmp = 0;
				if(flag_start == 1) buf1[analen++] = buf[i];
				else analen = 0;
			}
			if(cnt_cmp == sizeof(res_start) - 1) flag_start = 1;
			if(flag_start == 1 && buf[i] == '\n') {
				flag_end = 1;
				buf1[analen] = '\0';
				if(flag_end == 1) {
//					printf("%s\n", buf1);
					res_data.gid = ch2i(buf1[6]) * 16 + ch2i(buf1[7]); 
					res_data.rid = ch2i(buf1[15]) * 16 + ch2i(buf1[16]); 
					res_data.ch = ch2i(buf1[23]) * 16 + ch2i(buf1[24]); 
					for(j = 0; j < (int)sizeof(res_data.msg) - 1; j++) res_data.msg[j] = buf1[32 + j];
					res_data.idx = ch2i(buf1[63]) * 16 + ch2i(buf1[64]); 
					res_data.sid = ch2i(buf1[72]) * 16 + ch2i(buf1[73]); 
					for(j = 0; j < (int)sizeof(res_data.rt) - 1; j++) res_data.rt[j] = buf1[80 + j];

					printf("\n****************************************************\n");
					printf("%s\n", buf1);
					printf("GID(%d), RID(%d), CH(%d), IDX(%d), SID(%d)\n", res_data.gid, res_data.rid, res_data.ch, res_data.idx, res_data.sid);
					printf("Root Information: 0x%s\n", res_data.rt);

					/**** sample: analize sensor data: begin ****/
					if(res_data.msg[0] == '0' && res_data.msg[1] == '3') {			
						msg_03.unit_type[0] = res_data.msg[0];
						msg_03.unit_type[1] = res_data.msg[1];
						msg_03.ctrl_code[0] = res_data.msg[2];
						msg_03.ctrl_code[1] = res_data.msg[3];
						msg_03.battery[0] = res_data.msg[4];
						msg_03.battery[1] = res_data.msg[5];
						for(j = 0; j < 16; j++) msg_03.data[j] = res_data.msg[8 + j];
 						msg_03.data[j] = '\0';
						msg_03.tmp = ch2i(msg_03.data[2]) * 10 + ch2i(msg_03.data[3]) + ch2i(msg_03.data[4]) * 0.1;
						if(msg_03.data[1] == 1) msg_03.tmp = msg_03.tmp * (-1);
						msg_03.hum = ch2i(msg_03.data[6]) * 10 + ch2i(msg_03.data[7]) + ch2i(msg_03.data[8]) * 0.1;
						msg_03.lux = 0;
						for(j = 0; j < 5; j++) msg_03.lux = msg_03.lux * 10 + ch2i(msg_03.data[11 + j]);

						/**** output Tmp/Hum/Lux sensor data to stdout: begin ****/
						printf("\n--------- Tmp/Hum/Lux sensor data --------\n");
						printf("Temparature[degC]: %.1f\n", msg_03.tmp);
						printf("Humidity[%%RH]: %.1f\n", msg_03.hum);
						printf("Ambient[Lx]: %d\n", msg_03.lux);
						/**** output Tmp/Hum/Lux sensor data to stdout: end ****/

						/**** output Tmp/Hum/Lux sensor data to log file: begin ****/
						fp = fopen("log_tmp.txt", "w");
						fprintf(fp, "sid_tmp=%d&", res_data.sid);
						fprintf(fp, "tmp=%.1f&", msg_03.tmp);
						fprintf(fp, "hum=%.1f&", msg_03.hum);
						fprintf(fp, "lux=%d", msg_03.lux);
						fclose(fp);
						/**** output Tmp/Hum/Lux sensor data to log file: end ****/
					}
					else if(res_data.msg[0] == '0' && res_data.msg[1] == 'B') {
						msg_0b.unit_type[0] = res_data.msg[0];
						msg_0b.unit_type[1] = res_data.msg[1];
						msg_0b.ctrl_code[0] = res_data.msg[2];
						msg_0b.ctrl_code[1] = res_data.msg[3];
						msg_0b.battery[0] = res_data.msg[4];
						msg_0b.battery[1] = res_data.msg[5];
						for(j = 0; j < 12; j++) msg_0b.data[j] = res_data.msg[12 + j];
 						msg_0b.data[j] = '\0';
						msg_0b.detect_count = 0;
						for(j = 0; j < 12; j++) msg_0b.detect_count = msg_0b.detect_count * 10 + ch2i(msg_0b.data[j]);

						/**** output Human Detect sensor data to stdout: begin ****/
						printf("\n--------- Human Detect sensor data --------\n");
						printf("Count of Human Detction: %d\n", msg_0b.detect_count);
						/**** output Human Detect sensor data to stdout: end ****/

						/**** output Human Detect sensor data to log file: begin ****/
						fp = fopen("log_human.txt", "w");
						fprintf(fp, "sid_human=%d&", res_data.sid);
						fprintf(fp, "detect_count=%d", msg_0b.detect_count);
						fclose(fp);
						/**** output Human Detect sensor data to log file: end ****/
					}
					else if(res_data.msg[0] == '2' && res_data.msg[1] == '0') {
						msg_20.unit_type[0] = res_data.msg[0];
						msg_20.unit_type[1] = res_data.msg[1];
						msg_20.ctrl_code[0] = res_data.msg[2];
						msg_20.ctrl_code[1] = res_data.msg[3];
						msg_20.sub_code[0] = res_data.msg[4];
						msg_20.sub_code[1] = res_data.msg[5];
						for(j = 0; j < 12; j++) msg_20.data[j] = res_data.msg[12 + j];
 						msg_20.data[j] = '\0';
						msg_20.co2 = 0;
						for(j = 0; j < 12; j++) msg_20.co2 = msg_20.co2 * 10 + ch2i(msg_20.data[j]);

						/**** output CO2 sensor data to stdout: begin ****/
						printf("\n--------- CO2 sensor data --------\n");
						printf("CO2[ppm]: %d\n", msg_20.co2);
						/**** output CO2 sensor data to stdout: end ****/

						/**** output CO2 sensor data to log file: begin ****/
						fp = fopen("log_co2.txt", "w");
						fprintf(fp, "sid_co2=%d&", res_data.sid);
						fprintf(fp, "co2=%d", msg_20.co2);
						fclose(fp);
						/**** output CO2 sensor data to log file: end ****/
					}
					else {
						/**** reserve for the ohter node ****/
					}
					/**** sample: analize sensor data: end ****/
					flag_start = 0;
					flag_end = 0;
					cnt_cmp = 0;
					analen = 0;
				}
			}
		}
	}
/**** sample: receive data from nodes: end ****/

	return EXIT_SUCCESS;
}

